/*
  mod_telemetry - collect min, max, average runtimes for individual URI's

  John Adams <jna@retina.net>
  Retina Communications
  http://www.retina.net/tech
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Compile with:
   make; /gl/apache/bin/apxs -i -a -n telemetry libmodtelemetry.la; apachectl restart

  THIS MODULE REQUIRES THAT MOD_SCOREBOARD BE ENABLED. 

  Contains code derived from mod_bw (Ivan Barrera A.)
  http://bwmod.sourceforge.net

  ...and mod_status (Apache Group)
  http://www.apache.org
*/

/* conditional defines here */
#ifndef DEFAULT_TIME_FORMAT
#define DEFAULT_TIME_FORMAT "%A, %d-%b-%Y %H:%M:%S %Z"
#endif

#define TELEM_MAGIC_TYPE "application/x-httpd-telem"

#define MOD_COPYRIGHT_STRING "<I>mod_telemetry -- John Adams &lt;jna@retina.net&gt; -- <a href=\"http://code.google.com/p/modtelemetry/\">http://code.google.com/p/modtelemetry/</a></i>"

#define TELEM_DISABLED             1<<0
#define TELEM_ENABLED              1<<1

#define OPT_END       -1
#define OPT_RESET     0
#define OPT_SRT_URI   1
#define OPT_SRT_HITS  2
#define OPT_SRT_LASTDELTA  3
#define OPT_SRT_MIN   4
#define OPT_SRT_MAX   5
#define OPT_SRT_AVG   6
#define OPT_SRT_SLOW  7
#define OPT_SRT_LASTACC 8

/* scripts that take longer than these many microseconds = slow. */
#define SLOW_THRESHOLD1     (apr_time_t)1000
#define SLOW_THRESHOLD5     (apr_time_t)5000
#define SLOW_THRESHOLD10    (apr_time_t)10000

/* uncomment to enable detailed debugging */
#define DEBUG 1

/*
 * Include the core server components.
 */
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "apr_tables.h"
#include "apr_pools.h"
#include "apr_time.h"
#include "apr_hash.h"

#include "apr_atomic.h"
#include "apr_shm.h"
#include "time.h"

#include "scoreboard.h"

/* options structure */
struct telem_opt
{
    int id;
    const char *form_data_str;
    const char *hdr_out_str;
};

static const struct telem_opt telemetry_options[] =     /* see #defines above */
{
    {OPT_RESET, "reset", "Reset"},
    {OPT_SRT_URI, "uri", ""},
    {OPT_SRT_HITS, "hits", ""},
    {OPT_SRT_LASTDELTA, "last", ""},
    {OPT_SRT_MIN, "min", ""},
    {OPT_SRT_MAX, "max", ""},
    {OPT_SRT_AVG, "avg", ""},
    {OPT_SRT_SLOW, "slow", ""},
    {OPT_SRT_LASTACC, "lastacc", ""},
    {OPT_END, NULL, NULL}
};

/* this struct stores data about each unique URI */
typedef struct telemetry_s
{
    char uri[255];

    /* how many times it's violated each slow threshold */
    int slowsone;
    int slowsfive;
    int slowsten;

    /* times */
    apr_time_t min;
    apr_time_t max;
    apr_time_t avg;
    apr_time_t last;
    apr_time_t lastdelta;

    /* number of hits */
    apr_uint32_t hits;

    /* not using locking yet, should we? */
    volatile apr_uint32_t lock;

} telem_data;

/* brain-dead djb hash */
unsigned long djbhash(unsigned char *str)
{
    unsigned long hash = 5381;
    int c;

    while (c = *str++)
        hash = ((hash << 5) + hash) + c;        /* hash * 33 + c */

    return hash;
}


/* tmbase is shm base address */

/* initial support for 50000 uri's */
/* we'll need to make this automatic later on... */
static int sid = 50000;
telem_data *tmbase;
apr_shm_t *shm;
int sortby = OPT_SRT_MAX;

/* per server config */
typedef struct
{
    int state;                  /* enabled or disabled */
    int imgfilter;              /* enabled or disabled */
} telemetry_server_config;

module AP_MODULE_DECLARE_DATA telemetry_module;

/* Utility Functions */
static char *show_apr_time(apr_time_t tsecs)
{
  /* show an apr_time_t */ 
  int hrs, mins, secs;
  static char mystr[50];
  tsecs /= 1000;

  secs = (int) (tsecs % 60);
  tsecs /= 60;
  mins = (int) (tsecs % 60);
  tsecs /= 60;
  hrs = (int) (tsecs % 24);

  sprintf(mystr, "%02d:%02d:%02d.%02d",hrs,mins,secs);

  return mystr;
}

static void show_time(request_rec * r, apr_interval_time_t tsecs)
{
    int days, hrs, mins, secs;

    secs = (int) (tsecs % 60);
    tsecs /= 60;
    mins = (int) (tsecs % 60);
    tsecs /= 60;
    hrs = (int) (tsecs % 24);
    days = (int) (tsecs / 24);

    if (days)
        ap_rprintf(r, " %d day%s", days, days == 1 ? "" : "s");

    if (hrs)
        ap_rprintf(r, " %d hour%s", hrs, hrs == 1 ? "" : "s");

    if (mins)
        ap_rprintf(r, " %d minute%s", mins, mins == 1 ? "" : "s");

    if (secs)
        ap_rprintf(r, " %d second%s", secs, secs == 1 ? "" : "s");
}

char *stristr(const char *String, const char *Pattern)
{
    char *pptr, *sptr, *start;
    uint slen, plen;

    for (start = (char *) String,
         pptr = (char *) Pattern,
         slen = strlen(String), plen = strlen(Pattern);
         /* while string length not shorter than pattern length */
         slen >= plen; start++, slen--) {
        /* find start of pattern in string */
        while (toupper(*start) != toupper(*Pattern)) {
            start++;
            slen--;

            /* if pattern longer than string */

            if (slen < plen)
                return (NULL);
        }

        sptr = start;
        pptr = (char *) Pattern;

        while (toupper(*sptr) == toupper(*pptr)) {
            sptr++;
            pptr++;

            /* if end of pattern then pattern was found */

            if ('\0' == *pptr)
                return (start);
        }
    }
    return (NULL);
}

/*
 * This function is registered as a handler for HTTP methods and will
 * therefore be invoked for all GET requests (and others).  Regardless
 * of the request type, this function simply sends a message to
 * STDERR (which httpd redirects to logs/error_log).  A real module
 * would do *alot* more at this point.
 */
static int modtelemetry_log(request_rec * r)
{
    struct timeval start;
    const char *timestr;
    apr_time_t t1, t2;
    int loc;


    /* did this request fail? */
    if (!ap_is_HTTP_SUCCESS(r->status)) {
        return DECLINED;
    }

    /* filter out images */
    if ((stristr(r->uri, ".jpg") != NULL) ||
        (stristr(r->uri, ".png") != NULL) ||
        (stristr(r->uri, ".gif") != NULL)) {
        return DECLINED;
    }

    /* filter out other junk */
    if (stristr(r->uri, ".js") != NULL) {
        return DECLINED;
    }

    timestr = apr_table_get(r->notes, "tm_start");

    /* I don't know why strtoll works here and apr_atoi64 does not, but I am using strtoll.. */
    t1 = (apr_time_t) strtoll(timestr, NULL, 10);
    t2 = (apr_time_t) apr_time_now();

    apr_time_t delta = (apr_time_t) (t2 - t1);

    /* store data to SHM */
    loc = djbhash(r->uri) % sid;

    /* old entry and match, or new stuff? */
        if ((strcmp(tmbase[loc].uri, r->uri) == 0) ||
            (strcmp(tmbase[loc].uri, "") == 0)) {
      
#ifdef DEBUG
      fprintf(stderr, "%s start=%" APR_TIME_T_FMT " end=%" APR_TIME_T_FMT "\n",
        r->uri, t1, t2);
#endif
      strcpy(tmbase[loc].uri, r->uri);
      tmbase[loc].last = t1;

      /* is this a slow script? */
      if (apr_time_as_msec(delta) > SLOW_THRESHOLD1) {
        tmbase[loc].slowsone++;
      }
      if (apr_time_as_msec(delta) > SLOW_THRESHOLD5) {
        tmbase[loc].slowsfive++;
      }
      if (apr_time_as_msec(delta) > SLOW_THRESHOLD10) {
        tmbase[loc].slowsten++;
      }
      
      /* TODO: Handle Hash Collision. What to do?  */
      tmbase[loc].lastdelta = delta;
      tmbase[loc].hits++;
      
      /* moving avg, is this right? */
      if (tmbase[loc].avg == 0) {
        tmbase[loc].avg = delta;
      }
      else {
        tmbase[loc].avg = (tmbase[loc].avg + delta) / 2;
      }
      
      if (delta > tmbase[loc].max) {
        tmbase[loc].max = delta;
      };
      
      if ((delta < tmbase[loc].min) || (tmbase[loc].min == 0)) {
        tmbase[loc].min = delta;
      };

#ifdef DEBUG
    fprintf(stderr, "%s start=%" APR_TIME_T_FMT " end=%" APR_TIME_T_FMT
            " min=%" APR_TIME_T_FMT
            " max=%" APR_TIME_T_FMT
            " avg=%" APR_TIME_T_FMT
            "\n", r->uri, t1, t2, tmbase[loc].min, tmbase[loc].max,
            tmbase[loc].avg);
#endif

    fflush(stderr);
        } else {
#ifdef DEBUG
          fprintf(stderr,"mod_telemetry: hash collision - increase size of hash!!");
#endif
        }
        /* if existing or empty uri */
    return DECLINED;
}

static int reset_counters(apr_pool_t * p, server_rec * s)
{
    /* init all counters in shared memory */
    apr_status_t status;
    apr_size_t retsize;
    telem_data *tmstat;
    int t;

    /* Init APR's atomic functions */
    status = apr_atomic_init(p);
    if (status != APR_SUCCESS)
        return HTTP_INTERNAL_SERVER_ERROR;

    /* Init shm block */
    retsize = apr_shm_size_get(shm);
    tmbase = apr_shm_baseaddr_get(shm);
    if (tmbase == NULL) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                     "mod_telemetry : Error creating SHM  block.\n");
        return status;
    }
    memset(tmbase, 0, retsize);

    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                 "mod_telemetry : Memory Allocated %d bytes (each uri takes %d bytes)",
                 (int) retsize, (int) sizeof(telem_data));

    if (retsize < (sizeof(telem_data) * sid)) {
        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                     "mod_telemetry : Not enough memory allocated!! Giving up");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                 "mod_telemetry: Counters Reset\n");

}

/* sort functions */
static int cmp_maxtime(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].max < tmbase[j].max)
        return (1);
    if (tmbase[i].max > tmbase[j].max)
        return (-1);

    return (0);
}


static int cmp_mintime(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].min < tmbase[j].min)
        return (1);
    if (tmbase[i].min > tmbase[j].min)
        return (-1);

    return (0);
}

static int cmp_hits(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].hits < tmbase[j].hits)
        return (1);
    if (tmbase[i].hits > tmbase[j].hits)
        return (-1);

    return (0);
}

static int cmp_lastdelta(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].lastdelta < tmbase[j].lastdelta)
        return (1);
    if (tmbase[i].lastdelta > tmbase[j].lastdelta)
        return (-1);

    return (0);
}

static int cmp_lastacc(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].last < tmbase[j].last)
        return (1);
    if (tmbase[i].last > tmbase[j].last)
        return (-1);

    return (0);
}

static int cmp_avg(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].avg < tmbase[j].avg)
        return (1);
    if (tmbase[i].avg > tmbase[j].avg)
        return (-1);

    return (0);
}

static int cmp_slow(const void *p1, const void *p2)
{
    int i = *((int *) p1);
    int j = *((int *) p2);

    /* compare tmbase times */
    if (tmbase[i].slowsone < tmbase[j].slowsone)
        return (1);
    if (tmbase[i].slowsone > tmbase[j].slowsone)
        return (-1);

    return (0);
}



static int tm_report(request_rec * r)
{
    /* Dump the URI Report to the screen */
    int t;

    /* make an index of everything in memory */
    int index[sid];
    int indices = 0;
    int x;
    int totalrec = 0;
    apr_time_t nowtime;
    apr_interval_time_t up_time;

    nowtime = apr_time_now();

    for (x = 0; x < sid; x++) {
        if (tmbase[x].last > 0) {
            index[totalrec++] = x;
        }
    }

    /* shuffle the index based on sort criteria - default is by max time */
    switch (sortby) {
    case OPT_SRT_LASTDELTA:
        qsort((void *) index, totalrec, sizeof(int), cmp_lastdelta);
        break;
    case OPT_SRT_LASTACC:
        qsort((void *) index, totalrec, sizeof(int), cmp_lastacc);
        break;
    case OPT_SRT_HITS:
        qsort((void *) index, totalrec, sizeof(int), cmp_hits);
        break;
    case OPT_SRT_AVG:
        qsort((void *) index, totalrec, sizeof(int), cmp_avg);
        break;
    case OPT_SRT_SLOW:
        qsort((void *) index, totalrec, sizeof(int), cmp_slow);
        break;
    case OPT_SRT_MIN:
        qsort((void *) index, totalrec, sizeof(int), cmp_mintime);
        break;
    default:                   /* OPT_SRT_MAX */
        qsort((void *) index, totalrec, sizeof(int), cmp_maxtime);
        break;
    }

    /* display the page */

    ap_set_content_type(r, "text/html");
    ap_rputs(DOCTYPE_HTML_3_2
             "<html><head>\n<title>Telemetry</title>\n</head><body> \
<style type=\"text/css\"> \
body, td, p, div, small, big{ \
font-family: Arial, Helvetica, sans-serif \
} \
</style> ", r);
    /* server detail */
    up_time = (apr_uint32_t) apr_time_sec(nowtime -
                                          ap_scoreboard_image->global->
                                          restart_time);
    ap_rputs("<h1>Detailed Execution Time Report for ", r);
    ap_rvputs(r, ap_get_server_name(r), "</h1>\n\n", NULL);
    ap_rvputs(r, "<dl><dt>Server Version: ",
              ap_get_server_version(), "</dt>\n", NULL);
    ap_rvputs(r, "<dt>Server Built: ",
              ap_get_server_built(), "\n</dt></dl><hr /><dl>\n", NULL);
    ap_rvputs(r, "<dt>Current Time: ",
              ap_ht_time(r->pool, nowtime, DEFAULT_TIME_FORMAT, 0),
              "</dt>\n", NULL);
    ap_rvputs(r, "<dt>Restart Time: ",
              ap_ht_time(r->pool,
                         ap_scoreboard_image->global->restart_time,
                         DEFAULT_TIME_FORMAT, 0), "</dt>\n", NULL);
    ap_rprintf(r, "<dt>Parent Server Generation: %d</dt>\n",
               (int) ap_my_generation);
    ap_rputs("<dt>Server uptime: ", r);
    show_time(r, up_time);
    ap_rputs("</dt>\n", r);

    /* start our commands */
    ap_rputs("<P><a href=\"", r);
    ap_rputs(r->uri, r);
    ap_rputs("?reset\">[ Reset Counters ]</a>  | ", r);
    ap_rputs(" <a href=\"", r);
    ap_rputs(r->uri, r);

    /* retain argument list on refresh */
    if (r->args != NULL) {
        if (ap_strstr_c(r->args, "reset") == NULL) {
            ap_rputs("?", r);
            ap_rputs(r->args, r);
        }
    }
    ap_rputs("\">[ Refresh ] </a></p>", r);
    ap_rputs("<TABLE border=1 noshade>\n", r);

    ap_rputs
        ("<TR><TH><a href=\"/tm?uri\">URI</a></TH><TH><a href=\"/tm?hits\">Hits</a></TH><TH><a href=\"/tm?last\">Last Time</a></TH><TH><a href=\"/tm?min\">Min Time(mS)</a></TH><TH><a href=\"/tm?max\">Max Time(mS)</a></TH><TH><a href=\"/tm?avg\">Avg Time(mS)</a></TH><th><a href=\"/tm?slow\">Slow Count</a></th><TH><a href=\"/tm?lastacc\">Last Access</a></TH></TR>\n",
         r);

    /* display the list */
    for (t = 0; t < totalrec; t++) {
        if (tmbase[index[t]].last > 0) {
            char last_s[APR_CTIME_LEN];
            apr_ctime(last_s, tmbase[index[t]].last);

            ap_rputs((char *) apr_psprintf(r->pool,
                                           "<TR><TD>%s</TD><TD ALIGN=RIGHT>%d</TD><TD ALIGN=RIGHT>%"
                                           APR_TIME_T_FMT
                                           "</TD><TD ALIGN=RIGHT>%"
                                           APR_TIME_T_FMT
                                           "</TD><TD ALIGN=RIGHT>%"
                                           APR_TIME_T_FMT
                                           "</TD><TD ALIGN=RIGHT>%"
                                           APR_TIME_T_FMT
                                           "</TD><TD>%d/%d/%d (%.2f%%)</TD><TD>%s</TD></TR>\n",
                                           tmbase[index[t]].uri,
                                           tmbase[index[t]].hits,
                                           apr_time_as_msec(tmbase[index[t]].
                                                            lastdelta),
                                           apr_time_as_msec(tmbase[index[t]].
                                                            min),
                                           apr_time_as_msec(tmbase[index[t]].
                                                            max),
                                           apr_time_as_msec(tmbase[index[t]].
                                                            avg),
                                           tmbase[index[t]].slowsone,
                                           tmbase[index[t]].slowsfive,
                                           tmbase[index[t]].slowsten,
                                           (float) (100 *
                                                    ((float) tmbase[index[t]].
                                                     slowsone /
                                                     (float) tmbase[index[t]].
                                                     hits)), last_s), r);

        }
    }

    ap_rputs("</table><P>" MOD_COPYRIGHT_STRING "</P></body></html>\n", r);
    return OK;
}

static int modtelemetry_handle(request_rec * r)
{
    const char *timestr;
    struct timeval start;
    const char *loc;
    int i;

    apr_time_t t1 = apr_time_now();

    // store the start time in the module note field, we'll get this back later on.
    timestr = (char *) apr_psprintf(r->pool, "%" APR_TIME_T_FMT, t1);
    apr_table_set(r->notes, "tm_start", timestr);

#ifdef DEBUG
    fprintf(stderr, "req start: %s %s\n", r->uri, timestr);
    fflush(stderr);
#endif
    /* handle arguments */
    if (r->args) {
#ifdef DEBUG
        fprintf(stderr, "have args.");
        fflush(stderr);
#endif
        i = 0;
        while (telemetry_options[i].id != OPT_END) {
            if ((loc = ap_strstr_c(r->args,
                                   telemetry_options[i].form_data_str)) !=
                NULL) {
                switch (telemetry_options[i].id) {
                case OPT_RESET:
                    reset_counters(r->pool, r->server);
                    break;
                case OPT_SRT_URI:
                case OPT_SRT_HITS:
                case OPT_SRT_LASTDELTA:
                case OPT_SRT_MIN:
                case OPT_SRT_MAX:
                case OPT_SRT_AVG:
                case OPT_SRT_SLOW:
                case OPT_SRT_LASTACC:
                    /* change sort order */
                    sortby = telemetry_options[i].id;
                    break;
                }
            }
            i++;
        }
    }

    /* return OK if we're handling this... */
    if (strcmp(r->handler, "telemetry-status") == 0) {
        /* output the data */
        tm_report(r);
        return OK;
    }

    // Return DECLINED so that the Apache core will keep looking for
    // other modules to handle this request.  This effectively makes
    // this module completely transparent.

    return DECLINED;
}

/* Init The Module */
static int modtelemetry_post_config(apr_pool_t * p, apr_pool_t * plog,
                                    apr_pool_t * ptemp, server_rec * s)
{
    apr_status_t status;
    apr_size_t retsize;
    apr_size_t shm_size;
    telem_data *tmstat;
    int t;

    /* These two help ensure that we only init once. */
    void *data;
    const char *userdata_key = "ivn_shm_telem_module";

    apr_pool_userdata_get(&data, userdata_key, s->process->pool);

    if (!data) {
        apr_pool_userdata_set((const void *) 1, userdata_key,
                              apr_pool_cleanup_null, s->process->pool);
        return OK;
    }

    /* Init APR's atomic functions */
    status = apr_atomic_init(p);
    if (status != APR_SUCCESS)
        return HTTP_INTERNAL_SERVER_ERROR;

    shm_size = (apr_size_t) sizeof(telem_data) * sid;
    /* If there was a memory block already assigned.. destroy it */
    if (shm) {
        status = apr_shm_destroy(shm);
        if (status != APR_SUCCESS) {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                         "mod_telemetry : Couldn't destroy old memory block\n");
            return status;
        }
        else {
            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                         "mod_telemetry : Old Shared memory block, destroyed.");
        }
    }

    /* Create shared memory block */
    status = apr_shm_create(&shm, shm_size, NULL, p);
    if (status != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                     "mod_telemetry : Error creating shm block\n");
        return status;
    }
    /* Check size of shared memory block */
    retsize = apr_shm_size_get(shm);
    if (retsize != shm_size) {
        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
                     "mod_telemetry : Error allocating shared memory block\n");
        return status;
    }

    reset_counters(p, s);

    ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, s,
                 "mod_telemetry : Version %s - Initialized [%d uris]",
                 VERSION, sid);

    return OK;
}

/*
 * This function is a callback and it declares what other functions
 * should be called for request processing and configuration requests.
 * This callback function declares the Handlers for other events.
 */
static void modtelemetry_register_hooks(apr_pool_t * p)
{
    // I think this is the call to make to register a handler for method calls (GET PUT et. al.).
    // We will ask to be last so that the comment has a higher tendency to
    // go at the end.
    ap_hook_handler(modtelemetry_handle, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_post_config(modtelemetry_post_config, NULL, NULL,
                        APR_HOOK_MIDDLE);
    ap_hook_log_transaction(modtelemetry_log, NULL, NULL, APR_HOOK_LAST);

}

/* set mod enabled or disabled */
static const char *telemetrymodule(cmd_parms * cmd, void *dconf, int flag)
{
    telemetry_server_config *sconf;

    sconf =
        (telemetry_server_config *) ap_get_module_config(cmd->server->
                                                         module_config,
                                                         &telemetry_module);
    sconf->state = (flag ? TELEM_ENABLED : TELEM_DISABLED);

    return NULL;
}

static const char *telemetryfilter(cmd_parms * cmd, void *dconf, int flag)
{
    telemetry_server_config *sconf;

    sconf =
        (telemetry_server_config *) ap_get_module_config(cmd->server->
                                                         module_config,
                                                         &telemetry_module);
    sconf->imgfilter = (flag ? TELEM_ENABLED : TELEM_DISABLED);

    return NULL;
}


static void *create_tm_server_config(apr_pool_t * p, server_rec * s)
{
    /* allocate server config */
    telemetry_server_config *new;

    new =
        (telemetry_server_config *) apr_pcalloc(p,
                                                sizeof
                                                (telemetry_server_config));
    new->state = TELEM_DISABLED;
    new->imgfilter = TELEM_ENABLED;

    return (void *) new;
}

/*
 * Declare and populate the module's data structure.  The
 * name of this structure ('telem_module') is important - it
 * must match the name of the module.  This structure is the
 * only "glue" between the httpd core and the module.
 */


/* command table */
static const command_rec tm_cmds[] = {
    AP_INIT_FLAG("TelemetryModule", telemetrymodule, NULL,
                 RSRC_CONF | ACCESS_CONF,
                 "On or Off to enable or disable (default) the telemetry module"),
    AP_INIT_FLAG("TelemetryFilterImages", telemetryfilter, NULL,
                 RSRC_CONF | ACCESS_CONF,
                 "On or Off to enable (Default) or disable data collection for images"),
    {NULL}
};

module AP_MODULE_DECLARE_DATA telemetry_module = {
    // Only one callback function is provided.  Real
    // modules will need to declare callback functions for
    // server/directory configuration, configuration merging
    // and other tasks.
    STANDARD20_MODULE_STUFF,    /* Per dir config creator */
    NULL,                       /* dir config merger */
    NULL,                       /* dir merger */
    create_tm_server_config,    /* server config creator */
    NULL,                       /* merge svr config */
    tm_cmds,                    /* command table */
    modtelemetry_register_hooks,        /* callback for registering hooks */
};
